# 公开版代码
# 使用MNIST数据集进行手写字体数字的识别的demo，演示CNN方法进行分类的调用方法
# 本程序2024.04定型，后续更新都在完整版代码中
# 本代码地址及代码说明：http://khsci.com/docs/index.php/2024/04/21/cnnclasspy/
# 本代码知乎讲解：https://zhuanlan.zhihu.com/p/693677524
# 注意！本代码需要使用Python3.11版本，需要安装conda，环境配置请参考：https://zhuanlan.zhihu.com/p/693620785，以及本文件夹中的代码说明.docx
# 

import numpy as np
from khCNN import FunClassCNNs

# 1.加载MNIST数据集
data = np.load('mnist.npz')  # 从文件中加载MNIST数据集
xData, yData = data['x_train'], data['y_train']  # 获取训练集的特征和标签
xData = xData.reshape(-1, 1, 28, 28)  # 将训练集特征重塑为(样本数, 通道数, 高度, 宽度)的形状

# 2. 调用方法进行分类
# 2.1 设置数据集划分比例
divideR = [0.8, 0.1, 0.1]  # 训练集:验证集:测试集 = 8:1:1

# 2.2 设置卷积层参数
cLayer = np.array([
    [3, 3, 16, 1, 1],  # 第一个卷积层: 卷积核大小为3x3, 16个卷积核, 步长为1, 填充为1
    [3, 3, 32, 1, 0]   # 第二个卷积层: 卷积核大小为3x3, 32个卷积核, 步长为1, 填充为0
])

# 2.3 设置池化层参数
poolingLayer = [
    ['maxPooling2dLayer', 2, 2, 2, 0],  # 第一个池化层: 最大池化, 池化核大小为2x2, 步长为2, 填充为0
    ['averagePooling2dLayer', 2, 2, 2, 0]  # 第二个池化层: 平均池化, 池化核大小为2x2, 步长为2, 填充为0
]

# 2.4 设置全连接层参数
fcLayer = [128, 64]  # 两个全连接层, 输出维度分别为128和64

# 2.5 设置训练选项
options = {
    'MaxEpochs': 5,  # 最大迭代次数为5
    'MiniBatchSize': 128,  # 批量大小为128
    'InitialLearnRate': 0.001,  # 初始学习率为0.001
    'ValidationFrequency': 1,  # 每1次迭代进行一次验证
    'LearnRateSchedule': 'piecewise',  # 学习率调度方式为分段常数衰减
    'LearnRateDropPeriod': 10,  # 学习率下降周期为10
    'LearnRateDropFactor': 0.95  # 学习率下降因子为0.95
}

# 2.6 设置随机种子和设备选择
setting = {
    'seed': 42,  # 随机种子为42
    'deviceSel': 'gpu',  # 优先使用GPU进行训练
    'figflag': 'on',  # 绘制图像
    'minmax': True  # 进行归一化
}

# 2.7 调用函数进行训练和测试
FunClassCNNs(dataX=xData, dataY=yData, divideR=divideR,  cLayer=cLayer,poolingLayer=poolingLayer,fcLayer=fcLayer, options=options, setting=setting ) 

"""
def FunClassCNNs(dataX, dataY, divideR, cLayer, poolingLayer, fcLayer, options, setting):

    使用CNN进行模式识别(分类)的快速实现函数,程序会优先使用GPU进行加速,如果没有GPU则使用CPU
    
    参数:
    - dataX: 输入数据,形状为(num_samples, num_channels, height, width)的numpy数组
    - dataY: 标签值,形状为(num_samples,)的numpy数组,可以是向量型或索引型
    - divideR: 数据集划分比例,形如[train_ratio, val_ratio, test_ratio]的列表
    - cLayer: 卷积层结构,形状为(num_conv_layers, 5)的numpy数组,每一行代表一个卷积层的参数[filter_height, filter_width, num_filters, stride, padding]
    - poolingLayer: 池化层结构,形状为(num_conv_layers, 5)的列表,每一行代表一个池化层的参数['pool_type', pool_height, pool_width, stride, padding],其中pool_type可以是'maxPooling2dLayer'或'averagePooling2dLayer'或'none'
    - fcLayer: 全连接层结构,形状为(num_fc_layers,)的列表,每一个元素代表一个全连接层的输出维度,如果为空列表则只有一个输出维度等于类别数的全连接层
    - options: 网络训练相关的选项,字典类型,包含以下键值对:
      - 'solverName': 优化器类型,可以是'sgdm'或'rmsprop'或'adam',默认为'adam'
      - 'MaxEpochs': 最大迭代次数,默认为30
      - 'MiniBatchSize': 批量大小,默认为128
      - 'InitialLearnRate': 初始学习率,默认为0.005
      - 'ValidationFrequency': 验证频率,即每多少次迭代进行一次验证,默认为50
      - 'LearnRateSchedule': 学习率调度方式,可以是'piecewise'或'none',默认为'none'
      - 'LearnRateDropPeriod': 学习率下降周期,默认为10
      - 'LearnRateDropFactor': 学习率下降因子,默认为0.95
    - setting: 其他选项,字典类型,包含以下键值对:
      - figflag: 是否绘制图像,'on'为绘制,'off'为不绘制
      - deviceSel: 训练设备选择,可以是'cpu'或'gpu',默认为'gpu',当设置为'gpu'时,如果gpu硬件不可用,则会自动切换到cpu
      - seed: 随机种子,整数,设置为0时不启用,设置为其他整数时启用,不同的整数为不同的种子值,变换种子值会影响结果,相同种子的计算结果是一致的,缺省时为不设置随机种子
      - minmax: 是否进行归一化,布尔值,默认为True
      
    返回值:
    - accuracy: 测试集上的准确率
    - recall: 测试集上的召回率
    - precision: 测试集上的精确率
    - model: 训练好的PyTorch模型
    - info: 包含训练过程中的损失和准确率信息的字典
    """


# 关于完整版代码：
# 公开版代码的函数文件为pyd文件，可以被调用，但无法查看代码。
# 完整版代码中全部为py文件，py文件可以查看源码并自由修改，画图无水印
# 完整版可适用于一维、二维、三维数据（试用版仅支持二维数据）
# 完整版最大迭代次数无限制，可绘制网络结构图
# 完整版可导出预测结果、训练好的网络模型、神经网络训练相关参数（如loss值、准确度）等
# 购买完整版赠送CNN网络调参技巧视频教程

# 如果需要封装好的函数的源码，可在下述连接（完整版）获取：
# http://khsci.com/docs/index.php/2024/04/21/cnnclasspy/
# 编程不易，感谢支持~
